@@ -29,7 +29,13 @@ from transcriptic import routes
 from transcriptic.auth import StrateosSign
 from transcriptic.config import AnalysisException, Connection
 from transcriptic.english import AutoprotocolParser
-from transcriptic.util import ascii_encode, flatmap, iter_json, makedirs
+from transcriptic.util import (
+    PreviewParameters,
+    ascii_encode,
+    flatmap,
+    iter_json,
+    makedirs,
+)
 
 
 def submit(
@@ -846,17 +852,19 @@ def compile(protocol_name, args):
 
 
 def launch(
-    api,
-    protocol,
-    project,
-    title,
-    save_input,
-    local,
-    accept_quote,
-    params,
-    pm=None,
-    test=None,
-    pkg=None,
+    api: Connection,
+    protocol: str,
+    project: str,
+    title: str,
+    save_input: bool,
+    local: bool,
+    accept_quote: bool,
+    params: str,
+    pm: str = None,
+    test: bool = None,
+    pkg: str = None,
+    predecessor_id: str = None,
+    save_preview: bool = False,
 ):
     """Configure and launch a protocol either using the local manifest file or remotely.
     If no parameters are specified, uses the webapp to select the inputs."""
@@ -868,16 +876,26 @@ def launch(
             "`--payment` flag to use the default payment method."
         )
         return
-    # Load protocol from local file if not remote and load from listed protocols otherwise
-    if local:
-        manifest, protocol_obj = load_manifest_and_protocol(protocol)
-    else:
-        print_stderr(f"Searching for {protocol}...")
-        protocol_list = api.get_protocols()
 
+    # Project is required for quick launch
+    if not project:
+        click.echo(
+            "Project field is required if parameters file is not specified and is required for Run submission."
+        )
+        return
+    else:
+        project = get_project_id(api, project)
+        if not project:
+            return
+
+    # Load protocol from local file if not remote and load from listed protocols otherwise
+    if not local:
+        print_stderr(
+            f"Searching for {protocol} in organization {api.organization_id}..."
+        )
         matched_protocols = [
             p
-            for p in protocol_list
+            for p in api.get_protocols()
             if (p["name"] == protocol and (pkg is None or p["package_id"] == pkg))
         ]
 
@@ -893,53 +911,48 @@ def launch(
         else:
             print_stderr("Protocol found.")
         protocol_obj = matched_protocols[0]
+    else:
+        manifest, protocol_obj = load_manifest_and_protocol(protocol)
 
-    # If parameters are not specified, use quick launch to get inputs
+    # For remote execution, use input params file if specified, else use quick_launch inputs
     if not params:
-        # Project is required for quick launch
-        if not project:
-            click.echo("Project field is required if parameters file is not specified.")
-            return
-        project = get_project_id(api, project)
-        if not project:
-            return
-
+        # If parameters are not specified, use quick launch to get inputs
         # Creates web browser and generates inputs for quick_launch
         quick_launch = _get_quick_launch(api, protocol_obj, project)
+        params = dict(parameters=quick_launch["raw_inputs"])
+    else:
+        try:
+            params = json.loads(params.read())
+        except ValueError:
+            print_stderr(
+                "Unable to load parameters given. "
+                "File is probably incorrectly formatted."
+            )
+            return
 
-        # Save the protocol input locally if the user specified the save_input option
-        if save_input:
-            try:
-                with click.open_file(save_input, "w") as f:
-                    f.write(
-                        json.dumps(
-                            dict(parameters=quick_launch["raw_inputs"]), indent=2
-                        )
-                    )
-            except Exception as e:
-                print_stderr("\nUnable to save inputs: %s" % str(e))
+    # Save parameters to file if specified
+    if save_input:
+        try:
+            with click.open_file(save_input, "w") as f:
+                f.write(json.dumps(params, indent=2))
+        except Exception as e:
+            print_stderr("\nUnable to save inputs: %s" % str(e))
+
+    if save_preview:
+        pp = PreviewParameters(api, params["parameters"], protocol_obj)
+        # Read manifest.json and write updated manifest to working dir
+        try:
+            pp.merge(load_manifest())
+            with click.open_file("manifest.json", "w") as f:
+                f.write(json.dumps(pp.merged_manifest, indent=2))
+                f.close()
+        except Exception as e:
+            print_stderr(
+                f"\nUnable to save preview inputs due to not being"
+                f" able to process: {type(e)} {str(e)}"
+            )
 
     if not local:
-        # For remote execution, use input params file if specified, else use quick_launch inputs
-        if not params:
-            params = dict(parameters=quick_launch["raw_inputs"])
-            # Save parameters to file if specified
-            if save_input:
-                try:
-                    with click.open_file(save_input, "w") as f:
-                        f.write(json.dumps(params, indent=2))
-                except Exception as e:
-                    print_stderr("\nUnable to save inputs: %s" % str(e))
-        else:
-            try:
-                params = json.loads(params.read())
-            except ValueError:
-                print_stderr(
-                    "Unable to load parameters given. "
-                    "File is probably incorrectly formatted."
-                )
-                return
-
         req_id, launch_protocol = _get_launch_request(api, params, protocol_obj, test)
 
         # Check for generation errors
@@ -948,6 +961,21 @@ def launch(
         if len(generation_errs) > 0:
             for errors in generation_errs:
                 click.echo("\n\n" + str(errors["message"]))
+                if errors.get("info"):
+                    errors_info = errors.get("info")
+                    indexes = [
+                        idx
+                        for idx in range(len(errors_info))
+                        if errors_info.startswith("Error", idx)
+                        or errors_info.startswith("error", idx)
+                    ]
+                    # 100 length should give enough information
+                    errors_info_msgs = [
+                        str(errors_info[idx : idx + 100]) for idx in indexes
+                    ]
+                    for info_msg in errors_info_msgs:
+                        click.echo("\n" + info_msg)
+
             click.echo("\nPlease fix the above errors and try again.")
             return
 
@@ -964,14 +992,6 @@ def launch(
             if not confirmed:
                 return
 
-        # Project is required for run submission
-        if not project:
-            click.echo("\nProject field is required for run submission.")
-            return
-        project = get_project_id(api, project)
-        if not project:
-            return
-
         from time import gmtime, strftime
 
         if title:
@@ -987,9 +1007,12 @@ def launch(
                 title=run_title,
                 test_mode=test,
                 payment_method_id=pm,
+                predecessor_id=predecessor_id,
             )
             run_id = req_json["id"]
-            click.echo("\nRun created: %s" % api.url("%s/runs/%s" % (project, run_id)))
+            formatted_url = api.url(f"{project}/runs/{run_id}")
+            click.echo(f"\nRun created: {formatted_url}")
+            return formatted_url
         except Exception as err:
             click.echo("\n" + str(err))
     else:
@@ -1007,16 +1030,9 @@ def launch(
             In order to generate these `inputs`, we can create a new quick
             launch
             """
-            try:
-                params = json.loads(params.read())
-                # This is the input format required by resolve_inputs
-                formatted_inputs = dict(inputs=params["parameters"])
-            except ValueError:
-                print_stderr(
-                    "Unable to load parameters given. "
-                    "File is probably incorrectly formatted."
-                )
-                return
+            # This is the input format required by resolve_inputs
+            formatted_inputs = dict(inputs=params["parameters"])
+
             quick_launch = api.create_quick_launch(
                 data=json.dumps({"manifest": protocol_obj}), project_id=project
             )
@@ -1476,6 +1492,28 @@ def run_protocol(api, manifest, protocol, inputs, view=False, dye_test=False):
             return
 
 
+def validate_filter(filters, number_of_instructions):
+    invalid_filters = set()
+
+    for arg in filters:
+        if arg.isdigit():
+            idx = int(arg)
+            if 0 > idx or idx >= number_of_instructions:
+                invalid_filters.add(arg)
+        elif re.match(r"\d+-\d+", arg):
+            [s, e] = [int(v) for v in arg.split("-")]
+            if s > e or number_of_instructions <= e:
+                invalid_filters.add(arg)
+        elif ":" in arg:
+            tokens = arg.split(":")
+            if len(tokens) != 2:
+                invalid_filters.add(arg)
+        else:
+            invalid_filters.add(arg)
+
+    return invalid_filters
+
+
 def execute(
     autoprotocol,
     api,
@@ -1487,9 +1525,12 @@ def execute(
     schedule_at,
     schedule_delay,
     time_constraints_are_suggestion,
+    exclude,
+    include,
     partition_group_size,
     partition_horizon,
     partitioning_swap_device_id,
+    email,
 ):
     # Clean api end point
     if api.startswith("http://"):
@@ -1510,7 +1551,7 @@ def execute(
             "Error: '--schedule-delay' and '--schedule-at' are mutually exclusive.",
             err=True,
         )
-        return
+        sys.exit(1)
 
     # Get the requested scheduling time
     if schedule_delay is not None:
@@ -1520,12 +1561,27 @@ def execute(
         payload["scheduleAt"] = schedule_at
 
     # Get the autoprotocol
-    autoprotocol_str = autoprotocol.read()
     try:
-        payload["autoprotocol"] = json.loads(autoprotocol_str)
+        autoprotocol_json = json.loads(autoprotocol.read())
+        payload["autoprotocol"] = autoprotocol_json
     except json.decoder.JSONDecodeError as err:
         click.echo(f"Error decoding autoprotocol json: {err}", err=True)
-        return
+        sys.exit(1)
+
+    # validate filters
+    num_instructions = len(autoprotocol_json["instructions"])
+    invalid_exclude = validate_filter(exclude, num_instructions)
+    invalid_include = validate_filter(include, num_instructions)
+    if len(invalid_exclude) + len(invalid_include) > 0:
+        click.echo(
+            f"Error: invalid filters: {','.join(invalid_exclude.union(invalid_include))} (number of instructions: {num_instructions})",
+            err=True,
+        )
+        sys.exit(1)
+
+    # update the payload
+    payload["exclude"] = exclude
+    payload["include"] = include
 
     # device set resolution
     in_use = []
@@ -1536,7 +1592,7 @@ def execute(
             payload["deviceSet"] = device_json
         except json.decoder.JSONDecodeError as err:
             click.echo(f"Error decoding device set json: {err}", err=True)
-            return
+            sys.exit(1)
         in_use.append("--device-set")
 
     if workcell_id:
@@ -1551,7 +1607,7 @@ def execute(
 
     if len(in_use) > 1:
         click.echo(f"Error: {', '.join(in_use)} are mutually exclusive.", err=True)
-        return
+        sys.exit(1)
 
     if len(in_use) == 0:
         payload["workcellIdForDeviceSet"] = "wctest-mcx1"
@@ -1575,9 +1631,10 @@ def execute(
         path_tokens = clean_api.split("/")
         if len(path_tokens) != 3:
             click.echo(
-                f"Invalid api target, expects base-url/facility/workcell.", err=True
+                f"Error: Invalid api target, expects base-url/facility/workcell.",
+                err=True,
             )
-            return
+            sys.exit(1)
 
         clean_api = f"http://{clean_api}"
         path_base = f"http://{path_tokens[0]}"
@@ -1595,19 +1652,25 @@ def execute(
             ):
                 frontend_node_address = res_json["hostManifest"][path_lab][
                     path_workcell
-                ]
+                ]["url"]
             else:
                 click.echo(
-                    f"Error when get frontend node address: {res_json}", err=True
+                    f"Error when getting frontend node address: {res_json}", err=True
                 )
-                return
+                sys.exit(1)
         except json.decoder.JSONDecodeError:
-            click.echo(f"Error when get frontend node address: {res.text}", err=True)
-            return
+            click.echo(
+                f"Error when getting frontend node address: {res.text}", err=True
+            )
+            sys.exit(1)
+
+    # add sentBy
+    if email is not None:
+        payload["sentBy"] = email.split("@")[0]
 
     # POST to workcell
     test_run_endpoint = f"http://{frontend_node_address}/testRun"
-    click.echo(f"Sending request to http://{frontend_node_address}")
+    click.echo(f"Sending request to {test_run_endpoint}")
     res = requests.post(test_run_endpoint, json=payload)
     try:
         res_json = json.loads(res.text)
@@ -1615,6 +1678,8 @@ def execute(
             click.echo(
                 f"Success. View {clean_api}/dashboard to see the scheduling outcome."
             )
+            if "message" in res_json:
+                click.echo(res_json["message"])
         else:
             click.echo(f"Error: {res_json['message']}", err=True)
             if "sessionId" in res_json:
@@ -1622,8 +1687,10 @@ def execute(
                     f"Dashboard can be seen at: {clean_api}/dashboard",
                     err=True,
                 )
+                sys.exit(1)
     except json.decoder.JSONDecodeError:
         click.echo(f"Error: {res.text}", err=True)
+        sys.exit(1)
 
 
 def parse_json(json_file):
